from os import path
from collections import deque

class plink:
    """ Static class used to work with PLINK-formatted data
    """
    @staticmethod
    def write_geno_mat_to_ped(write_base, geno_mat, samples, positions,
                              chrom, families=None):
        """ Writes MAP and PED files for a given set of genotypes

        Parameters
        ----------
        write_base : str
            The base name of the resulting plink files
        geno_mat : np.ndarray
            A numpy matrix of integer genotypes (0,1,2). All markers are assumed
            to be biallelic. Rows are samples,  and columns are positions
        samples : iterable
            The sample names, in the same order as matrix rows
        positions : iterable
            The positions of the markers, in the same order as matrix columns
        chrom : str
            The chromosome name
        families : iterable, optional
            Family name for each individual, if they are grouped by family
        """
        # First write the MAP file
        map_file = open(write_base+'.map','w')
        for i,position in enumerate(positions):
            line = [chrom,"marker%d" % i,"0",str(position)]
            map_file.write(' '.join(line)+'\n')
        map_file.close()
        # Write the ped file
        ped_file = open(write_base+'.ped','w')
        for i,samp in enumerate(samples):
            if families:
                write_line = deque([families[i],samp,'0','0','0','0'])
            else:
                write_line = deque([str(i+1),samp,'0','0','0','0'])
            for j,pos in enumerate(positions):
                if geno_mat[i,j] == 0:
                    write_line.append('1')
                    write_line.append('1')
                elif geno_mat[i,j] == 1:
                    write_line.append('1')
                    write_line.append('2')
                elif geno_mat[i,j] == 2:
                    write_line.append('2')
                    write_line.append('2')
            ped_file.write(' '.join(write_line)+'\n')
        ped_file.close()
        



import gzip
import re
import numpy as np
from Ranger import RangeSet
from genomfart.utils.caching import LRUcache
from bisect import bisect_left

## Compile regular expressions
int_re = re.compile(r'^-*[\d]+$')
float_re = re.compile(r'(^-*[\d]+\.[\d]+$|^-*[\d]{1}([\.]?[\d])*(E|e)-[\d]+$)')

class BigDataFrame(object):
    """ Class acting as a DataFrame for large datasets

    Not all of the dataset is loaded into memory. Rather, a cache of rows is
    maintained along with a handle and a position in the file
    """
    def __init__(self, filename, header = True, sep='\t',
                 detect_type = True, assume_uniform_types = True,
                 ignore_chars = None, maxsize=10000,
                 byte_record_interval=10000):
        """ Instantiates the big data frame

        Parameters
        ----------
        filename : str
            The file containing delimited data
        header : boolean, optional
            Whether the file has a header
        rowLabs : int, optional
            0-based column containing headers for the rows
        sep : str, optional
            The separateor of columns in the file (Can be a compiled regular expression)
        detect_type : boolean, optional
            Whether type should be guessed on loading. If false, everythin returned
            as str
        assume_uniform_types : boolean, optional
            Whether it should be assumed that the types in the first row apply to all rows.
            Only matters if detect_type is True
        ignore_chars : Set
            Characters that signal a line in the file should be skipped if it
            starts with one of those characters
        maxsize : int, optional
            The maximum number of rows to cache
        byte_record_interval : int, optional
            How often to record the starting byte of a row
        """
        self.byte_record_interval = byte_record_interval
        self.ignore_chars = ignore_chars if ignore_chars else set()
        ## Dictionary with headers for each of the columns, where values are integer
        # values
        self.colLabels = {}
        ## Dictionary with headers for each of the rows, where values are integer
        # values
        self.rowLabels = {}
        ## LRU cache holding rows of data, with row indices as keys
        self.data = LRUcache(self._fetch_from_cache, maxsize = maxsize)
        ## Function used to convert to correct type
        self.type_func = lambda x: x[1]
        self.type_dict = {}
        if detect_type:
            if assume_uniform_types:
                self.type_func = lambda x: self.type_dict[x[0]](x[1])
            else:
                def type_func(x):
                    if float_re.match(x[1]): return float(x[1])
                    elif int_re.match(x[1]): return int(x[1])
                    else:
                        return str(x[1])
                self.type_func = type_func
        ## Function used to split lines
        if type(sep) == type(int_re):
            self.split_func = lambda x: re.split(sep, x)
        elif type(sep) == str:
            self.split_func = lambda x: x.split(sep)
        else:
            raise TypeError("Sep must be compiled regular expression or string")
        ## Get the file handle
        if filename.endswith('.gz'):
            self.handle = gzip.open(filename)
        else:
            self.handle = open(filename)
        ## Skip anything that needs to be skipped
        current_byte = self.handle.tell()
        while 1:
            current_byte = self.handle.tell()
            line = self.handle.readline().strip()
            if line[0] in self.ignore_chars: continue
            else:
                break
        self.handle.seek(current_byte)
        ## Incorporate header if necessary
        if header:
            line = self.handle.readline().strip()
            if ignore_chars:
                while 1:
                    if line[0] in self.ignore_chars:
                        line = self.handle.readline().strip()
                    else:
                        break
            line = self.split_func(line)
            for i,colname in enumerate(line):
                self.colLabels[colname] = i
        ## Create dictionary of line index -> byte
        self.row_byte_dict = {0: self.handle.tell()}                
        ## Get the types
        line = self.handle.readline().strip()
        line = self.split_func(line)
        for i, val in enumerate(line):
            if float_re.match(val):
                self.type_dict[i] = float
            elif int_re.match(val):
                self.type_dict[i] = int
            else:
                self.type_dict[i] = str
        self.handle.seek(self.row_byte_dict[0])
        ## Ordered list of the line indices that have been recorded
        self.ordered_row_inds = [0]
        ## The current line index
        self.current_line_ind = 0
    def __getitem__(self, inds):
        """ Gets an entry or a row from the data frame

        Parameters
        ----------
        inds : int or tuple
            Either the zero-based index of a row to get or a tuple of
            two numbers, where the first is the row and the second is the column,
            or a tuple of a number and a string, where the string is the name
            of the column to get

        Returns
        -------
        Either a single entry or a dictionary for the row
        """
        if type(inds) == tuple:
            inds = list(inds)
            if len(inds) > 2:
                raise TypeError("Two indices maximum")
            elif type(inds[1]) == str:
                # Get the index associated with the column
                try:
                    inds[1] = self.colLabels[inds[1]]
                except KeyError:
                    raise KeyError("%s is not a column label" % inds[1])
            if type(inds[0]) != int:
                raise TypeError("Row index must be integer")
            # Get from the cache
            row = self._fetch_from_cache(inds[0])
            return row[inds[1]]
        elif type(inds) == int:
            row = self._fetch_from_cache(inds)
            # Return the row as a dictionary
            if len(self.colLabels) > 0:
                return dict((k,row[v]) for k,v in self.colLabels.items())
            else:
                return dict((i,x) for i,x in enumerate(row))
        else:
            raise TypeError("Row index must be integer")
    def __iter__(self):
        """ Iterates through the rows of the data frame. Will cache as it
        moves through the rows

        Returns
        -------
        Generator of dictionaries for each row
        """
        self.current_line_ind = self._goto_closest_line(0)
        header = True if len(self.colLabels) > 0 else False
        for line in self.handle:
            row = tuple(map(self.type_func, enumerate(self.split_func(line.strip()))))
            self.data[self.current_line_ind] = row
            if header:
                yield dict((k,row[v]) for k,v in self.colLabels.items())
            else:
                yield dict((i,x) for i,x in enumerate(row))            
            self.current_line_ind += 1
    def _fetch_from_cache(self, row_ind):
        """ Gets a row from the cache if present. Otherwise, puts it in the cache
        and fetches it

        Parameters
        ----------
        row_ind : int
            The index of the row

        Returns
        -------
        The row as a tuple
        """
        if row_ind in self.data:
            return self.data[row_ind]
        elif self.current_line_ind == row_ind:
            line = self.handle.readline()
            if (self.current_line_ind + 1) % self.byte_record_interval == 0:
                # Record the current byte if necessary
                self.row_byte_dict[self.current_line_ind+1] = self.handle.tell()
            if self.current_line_ind == row_ind:
                ## If we're on the line we want
                # Put the line in the cache after processing
                self.data[self.current_line_ind] = tuple(map(self.type_func,
                                                enumerate(self.split_func(line.strip()))))
                return self.data[self.current_line_ind]           
        else:
            # Send the handle to the closest line
            self.current_line_ind = self._goto_closest_line(row_ind)
            for line in self.handle:
                if (self.current_line_ind + 1) % self.byte_record_interval == 0:
                    # Record the current byte if necessary
                    self.row_byte_dict[self.current_line_ind+1] = self.handle.tell()
                if self.current_line_ind == row_ind:
                    ## If we're on the line we want
                    # Put the line in the cache after processing
                    self.data[self.current_line_ind] = tuple(map(self.type_func,
                                                    enumerate(self.split_func(line.strip()))))
                    return self.data[self.current_line_ind]
                self.current_line_ind += 1
        # If nothing else, there's an index error
        raise IndexError("Line %d is not in the file" % row_ind)
            
    def _goto_closest_line(self, line_ind):
        """ Sends the handle to the line closest to (but before) the one you
        want that is already recorded in the row_byte_dict

        Parameters
        ----------
        line_ind : int
            The line you want to go to

        Returns
        -------
        The 0-based index of the row where the handle currently sits
        """
        closest_ind_ind = max(bisect_left(self.ordered_row_inds, line_ind)-1, 0)
        # Send the handle to the location
        self.handle.seek(self.row_byte_dict[closest_ind_ind])
        self.current_line_ind = self.ordered_row_inds[closest_ind_ind]
        return self.ordered_row_inds[closest_ind_ind]
    def get_current_row_ind(self):
        """ Gets the index of the current row

        Returns
        -------
        The zero-based index of the current row
        """
        return self.current_line_ind
    def iterrows(self, cache=False, colspec=None):
        """ Iterates through rows in the dataframe, with the option of whether
        or not to cache rows while moving through

        Parameters
        ----------
        cache : boolean
            Whether or not to cache rows while iterating. Choosing False
            will result in a performance increase while iterrating, but O(1) access
            to iterrated rows will not be available afterward
        colspec : iterable
            Particular columns (index, or name if there is a header)
            that should be included. If given, all other columns will be
            discarded

        Returns
        -------
        Generator of dictionaries for each row        
        """
        self.current_line_ind = self._goto_closest_line(0)
        header = True if len(self.colLabels) > 0 else False
        for line in self.handle:
            if colspec:
                row = self.split_func(line.strip())
                if cache:
                    self.data[self.current_line_ind] = tuple(map(self.type_func, enumerate(row)))
                if header:
                    yield dict((k, self.type_func((self.colLabels[k],row[self.colLabels[k]]))) for \
                               k in colspec)
                else:
                    yield dict((i, self.type_func((i, row[i]))) for i in colspec)
            else:
                row = tuple(map(self.type_func, enumerate(self.split_func(line.strip()))))
                if cache:
                    self.data[self.current_line_ind] = row
                if header:
                    yield dict((k,row[v]) for k,v in self.colLabels.items())
                else:
                    yield dict((i,x) for i,x in enumerate(row))            
            self.current_line_ind += 1
    def make_numpy_array(self, rows = None, cols=None):
        """ Create a 2-dimensional numpy array of data in the data frame.

        Parameters
        ----------
        rows : list of ints or Ranger.RangeSet
           Either a list of rows to include in the array or a RangeSet
           that includes all row indices to be included in the RangeSet. If None,
           all rows will be included
        cols : list of ints or Ranger.RangeSet
           Either a list of columns to include in the array or a RangeSet
           that includes all columns to be included in the RangeSet. If None,
           all columns will be included

        Returns
        -------
        A numpy array containing the data
        """
        ## Create function that deterimines whether the row RangeSet
        # or list contains a given row index
        row_contain_func = lambda x: True
        if rows and any(list(map(lambda x: isinstance(rows, x), (set, frozenset,
                                                                 list, tuple)))):
            rows = frozenset(rows)
            row_contain_func = lambda x: x in rows
        elif rows and isinstance(rows, RangeSet):
            row_contain_func = lambda x: rows.contains(x)
        elif rows:
            raise TypeError("Rows must be specified as a set, tuple, list, or RangeSet")
        # Make a list of the columns to get
        if cols is None:
            cols = range(len(self[0]))
        elif isinstance(cols, RangeSet):
            colsList = []
            for subRange in cols:
                rangeStart = subRange.lowerEndpoint() if subRange.isLowerBoundClosed() \
                  else subRange.lowerEndpoint()+1
                rangeEnd = subRange.upperEndpoint() if not subRange.isUpperBoundClosed() \
                  else subRange.upperEndpoint()+1
                colsList += range(rangeStart, rangeEnd)
            cols = colsList
        elif not any(list(map(lambda x: isinstance(cols, x), (list, tuple)))):
            raise TypeError("Cols must be specified as a RangeSet, tuple, or list")
        # Instantiate a list to holds to rows of the resulting array
        returnArr = []
        # Iterate through the rows, adding to the array
        for i,row in enumerate(self):
            if row_contain_func(i):
                returnArr.append(list(map(lambda x: self[i,x], cols)))
            else:
                continue
        # Return the array
        return np.array(returnArr)



import networkx as nx
from Ranger import RangeBucketMap, Range

class genomeAnnotationGraph(object):
    """ Representation of a genome, in which annotations in the genome are Ranges
    in a RangeBucketMap, and annotations can be hierarchical (e.g. transcripts are the children
    of genes). The hierarchy is stored as a directed graph
    """
    def __init__(self):
        """ Instantiates the genomeAnnotationGraph
        """
        ## Set up a directed graph where nodes are ids and a directed edge
        # is placed going from parent to child. Each node has a "Ranges" attribute
        # that lists the Ranges corresponding to the ID, a "seqid" attribute
        # that gives the coordinate system where the node is located, a
        # "type" attribute that gives the element type, a "strand" attribute
        # that gives the element strand (not always applicable), and an "attributes"
        # attribute that is a list of dictionaries of key-> val for any other attributes        
        self.graph = nx.DiGraph()
        ## Dictionary of seq_id -> RangeBucketMap for each coordinate system. Each Range
        # in a RangeBucketMap maps to an ID that corresponds to a node in the graph
        self.bucketmaps = {}
    def add_annotation(self, element_id, seqid, start, end, element_type,
                       strand='?', parents = None, children = None,  **attr):
        """ Adds an annotation to the genome

        Note that you can add more than one annotation under an element id,
        so this can be preexisting. If it is preexisting, the Range and attribute dictionaries
        will be appended

        Parameters
        ----------
        element_id : hashable
            The id of the element you're adding. 
        seqid : hashable
            The name of the coordinate system (e.g. the chromosome)
        start : int
            The start of the element (inclusive)
        end : int
            The end of the element (inclusive)
        element_type : str
            The type of the element (e.g. 'gene' or 'mRNA')
        strand : str, optional
            The element strand ('+','-','?')
        parents : iterable of strings, optional
            The parent element_ids of this element_id
        children : iterable of strings, optional
            The child element_ids of this element_id
        attr : Further keyword arguments that will be stored as attributes
               of the element
        """
        if not parents: parents = set()
        if not children: children = set()
        # Put the seqid in the bucketmaps dictionary if necessary
        if seqid not in self.bucketmaps:
            self.bucketmaps[seqid] = RangeBucketMap()
        # Make the Range for this element
        element_range = Range.closed(start, end)
        # Put element in RangeBucketMap
        self.bucketmaps[seqid].put(element_range, element_id)
        # Put node in graph if necessary
        if element_id not in self.graph:
            self.graph.add_node(element_id)
            self.graph.node[element_id]['Ranges'] = []
            self.graph.node[element_id]['attributes'] = []
        # Add/update parameters in graph
        self.graph.node[element_id]['seqid'] = seqid
        self.graph.node[element_id]['Ranges'].append(element_range)
        self.graph.node[element_id]['attributes'].append(attr)
        self.graph.node[element_id]['type'] = element_type
        self.graph.node[element_id]['strand'] = strand
        # Make edges if necessary
        for parent in parents:
            self.graph.add_edge(parent, element_id)
        for child in children:
            self.graph.add_edge(element_id, child)
    def get_aa_indices(self, seqid, pos, cds_type = 'CDS'):
        """ Gets the indices (base-1) of the amino acid position in any
        CDS overlapping it

        Parameters
        ----------
        seqid : str
            The name of the coordinate system to check
        pos : int
            The position to check (1-based)
        cds_type : str, optional
            The element type containing the CDS ranges

        Returns
        -------
        Dictionary of CDS_ID -> AA position in protein (1-based)
        """
        return_dict = {}
        # Get CDS position(s)
        cds_inds = self.get_cds_indices(seqid, pos, cds_type)
        for k,v in cds_inds.items():
            return_dict[k] = int((v-1)/3)+1
        return return_dict
    def get_cds_indices(self, seqid, pos, cds_type='CDS'):
        """ Gets the indices (base-1) of the nucleotide position in any
        CDS overlapping it

        Parameters
        ----------
        seqid : str
            The name of the coordinate system to check
        pos : int
            The position to check (1-based)
        cds_type : str, optional
            The element type containing the CDS ranges

        Returns
        -------
        Dictionary of CDS_ID -> position in CDS (1-based)
        """
        return_dict = {}
        # Get CDS overlapping the position
        cds_ids = self.get_overlapping_element_ids_of_type(seqid, pos,
                                                           pos, cds_type)
        # Go through CDS getting the position
        for cds_id in cds_ids:
            add_value = 0
            cds_ranges = sorted(self.graph.node[cds_id]['Ranges'],
                                key = lambda x: x.lowerEndpoint() if \
                                self.graph.node[cds_id]['strand'] == '+' else \
                                -x.lowerEndpoint())
            for cds_range in cds_ranges:
                if cds_range.contains(pos):
                    if self.graph.node[cds_id]['strand'] == '+':
                        return_dict[cds_id] = add_value + \
                          (pos-cds_range.lowerEndpoint()+1)
                        break
                    elif self.graph.node[cds_id]['strand'] == '-':
                        return_dict[cds_id] = add_value + \
                          (cds_range.upperEndpoint()-pos+1)
                        break
                else:
                    add_value += (cds_range.upperEndpoint()-cds_range.lowerEndpoint()+1)
        return return_dict
    def get_codon_position(self, seqid, pos, cds_type='CDS'):
        """ Gets the indices (base-1) of the codon position (i.e. 1,2, or 3)
        within any CDS 

        Parameters
        ----------
        seqid : str
            The name of the coordinate system to check
        pos : int
            The position to check (1-based)
        cds_type : str, optional
            The element type containing the CDS ranges

        Returns
        -------
        Dictionary of CDS_ID -> codon position
        """
        return_dict = {}
        # Get CDS position(s)
        cds_inds = self.get_cds_indices(seqid, pos, cds_type)
        for k,v in cds_inds.items():
            return_dict[k] = ((v-1) % 3)+1
        return return_dict        
    def add_node_annotations(self, element_id, **annots):
        """ Adds annotations to all annotation dictionaries under a node.
        Note that if the key for an annotation is the same as a previously
        existing one, that annotation will be overwritten

        Parameters
        ----------
        element_id : hashable
            The id of the element you're adding.
        annots : keyword arguments
            The elements you want to add

        Examples
        --------

        >>> my_genome.add_node_annotations('myNode1', expression1 = 0.2, expression2 = 10.)
        """
        for attr_dict in self.graph.node[element_id]['attributes']:
            attr_dict.update(annots)
    def get_overlapping_element_ids(self, seqid, start, end):
        """ Gets the ids for any elements that overlap a given range

        Parameters
        ----------
        seqid : str
            The name of the coordinate system to check
        start : int
            The start of the range to check (inclusive, 1-based)
        end : int
            The end of the range to check (inclusive, 1-based)

        Raises
        ------
        KeyError
            If the seqid is not present
        
        Returns
        -------
        Set of ids for elements overlapping the range
        """
        if seqid not in self.bucketmaps:
            raise KeyError("%s not present" % seqid)
        checkRange = Range.closed(start, end)
        try:
            return self.bucketmaps[seqid].get(checkRange)
        except KeyError:
            return set()
    def get_overlapping_element_ids_of_type(self, seqid, start, end, element_type):
        """ Gets the ids for any elements that overlap a given range and are
        of a given type

        Parameters
        ----------
        seqid : str
            The name of the coordinate system to check
        start : int
            The start of the range to check (inclusive, 1-based)
        end : int
            The end of the range to check (inclusive, 1-based)
        element_type : str
            The element type that you want (e.g. 'gene', 'mRNA')

        Raises
        ------
        KeyError
            If the seqid is not present

        Returns
        -------
        Set of ids for elements overlapping the range that are of the
        element type
        """
        all_overlaps = self.get_overlapping_element_ids(seqid, start, end)
        return set([x for x in all_overlaps if self.graph.node[x]['type'] == element_type])
    def get_element_info(self, element_id):
        """ Gets information on a particular element

        Parameters
        ----------
        element_id : str
            The id of the element

        Returns
        -------
        Dictionary of {'seqid' -> seqid, type'->type, 'strand'->'strand', 'Ranges'->[Ranges],
        'attributes'->[attribute_dicts]}
        """
        return {'type':self.graph.node[element_id]['type'],
                'strand':self.graph.node[element_id]['strand'],
                'seqid':self.graph.node[element_id]['seqid'],
                'Ranges':self.graph.node[element_id]['Ranges'],
                'attributes':self.graph.node[element_id]['attributes']}
    def get_element_ids_of_type(self, seqid, element_type, start = None, end = None):
        """ Gets element ids of some type along a coordinate system

        Parameters
        ----------
        seqid : str
            The name of the coordinate system to check
        element_type : str
            The type of the elements you want (e.g. 'gene' or 'mRNA')
        start : int, optional
            The start point for getting the elements (inclusive, 1-based)
        end : int, optional
            The end point for getting the elements (inclusive, 1-based)

        Raises
        ------
        KeyError
            If the seqid isn't included
        
        Returns
        -------
        Generator of element_ids
        """
        added = set()
        iterator = self.bucketmaps[seqid].iteritems(start=start,end=end)
        for theRange, element_id in iterator:
            if element_id in added:
                continue
            elif self.graph.node[element_id]['type'] == element_type:
                yield element_id
                added.add(element_id)
    def get_closest_element_id(self, seqid, rangeStart, rangeEnd, radius = 10000):
        """ Gets the element id(s) of the whatever element is closest to a range

        Parameters
        ----------
        seqid : str
            The name of the coordinate system to check
        rangeStart : int
            The position (inclusive) beginning the range for which you want
            the closest element
        rangeEnd : int
            The position (inclusive) ending the range for which you want
            the closest element
        radius : int
            How far on either side of the search range you want to search
            for the closest element

        Returns
        -------
        Set that will contain all element ids overlapping the search range
        if there are any. Otherwise, it will contain the element(s) with the
        shortest distance to the search range (only more than 1 if some elements
        are equidistant)
        """
        checkRange = Range.closed(rangeStart, rangeEnd)
        # Go through items overlapping the Range
        iterator = self.bucketmaps[seqid].iteritems(start=(rangeStart-radius),
                                                    end=(rangeEnd+radius))
        # Store the distance and element id of closest element that is not
        # overlapping the search range and is to the left of the search range
        less_than_closest_dist = radius
        less_than_closest_element = set()
        # Store the distance and element id of the closest element that is
        # not overlapping the search range and is to the right of the search range
        greater_than_closest_dist = radius
        greater_than_closest_element = set()
        # Store any overlapping elements
        overlapping = set()
        for theRange, element_id in iterator:
            # Close the Range if necessary
            if not all((theRange.isLowerBoundClosed(), theRange.isUpperBoundClosed())):
                newLowerBd = theRange.lowerEndpoint() if theRange.isLowerBoundClosed() else \
                  (theRange.lowerEndpoint()+1)
                newUpperBd = theRange.upperEndpoint() if theRange.isUpperBoundClosed() else \
                  (theRange.upperEndpoint()-1)
                theRange = Range.closed(newLowerBd, newUpperBd)
            if checkRange.isConnected(theRange):
                # Add element to overlapping set if overlapping
                overlapping.add(element_id)
            elif theRange.lowerEndpoint() < checkRange.lowerEndpoint():
                # If the element is to the left of the search range
                range_dist = checkRange.getDistanceFromRange(theRange)
                if range_dist < less_than_closest_dist:
                    # Add the element to the less than set after clearing current
                    # contents 
                    less_than_closest_element.clear()
                    less_than_closest_element.add(element_id)
                    less_than_closest_dist = range_dist
                elif range_dist == less_than_closest_dist:
                    # Add the element to the current set
                    less_than_closest_element.add(element_id)
            elif theRange.lowerEndpoint() > checkRange.upperEndpoint():
                # If the element is to the right of the search range
                if len(overlapping) > 0:
                    # Return the overlapping set if there are any overlapping elements
                    return overlapping
                range_dist = checkRange.getDistanceFromRange(theRange)
                if range_dist > less_than_closest_dist:
                    # Return the closest elements if the current range distance is
                    # greater than the left distances
                    if less_than_closest_dist < greater_than_closest_dist:
                        return less_than_closest_element
                    elif greater_than_closest_dist < less_than_closest_dist:
                        return greater_than_closest_element
                    else:
                        return less_than_closest_element.union(greater_than_closest_element)
                elif range_dist < greater_than_closest_dist:
                    # Add the element to the greater than set after clearing
                    # current contents
                    greater_than_closest_element.clear()
                    greater_than_closest_element.add(element_id)
                    greater_than_closest_dist = range_dist
                elif range_dist == greater_than_closest_dist:
                    # Add the element to the current set
                    greater_than_closest_element.add(element_id)
                elif range_dist > greater_than_closest_dist:
                    # Return the closest elements if the current range distance
                    # is greater than the minimum right distance
                    if less_than_closest_dist < greater_than_closest_dist:
                        return less_than_closest_element
                    elif greater_than_closest_dist < less_than_closest_dist:
                        return greater_than_closest_element
                    else:
                        return less_than_closest_element.union(greater_than_closest_element)
        # Return the closest if finished going through all elements
        if len(overlapping) > 0:
            return overlapping
        elif less_than_closest_dist < greater_than_closest_dist:
            return less_than_closest_element
        elif greater_than_closest_dist < less_than_closest_dist:
            return greater_than_closest_element
        elif less_than_closest_dist == greater_than_closest_dist:
            return less_than_closest_element.union(greater_than_closest_element)

    def get_closest_element_id_of_type(self, seqid, rangeStart, rangeEnd, element_type,
                                       radius = 10000):
        """ Gets the element id(s) of the whatever element is closest to a range

        Parameters
        ----------
        seqid : str
            The name of the coordinate system to check
        rangeStart : int
            The position (inclusive) beginning the range for which you want
            the closest element
        rangeEnd : int
            The position (inclusive) ending the range for which you want
            the closest element
        radius : int
            How far on either side of the search range you want to search
            for the closest element
        element_type : str
            The type of the elements you want (e.g. 'gene' or 'mRNA')

        Returns
        -------
        Set that will contain all element ids overlapping the search range
        if there are any. Otherwise, it will contain the element(s) with the
        shortest distance to the search range (only more than 1 if some elements
        are equidistant)
        """
        checkRange = Range.closed(rangeStart, rangeEnd)
        # Go through items overlapping the Range
        iterator = self.bucketmaps[seqid].iteritems(start=(rangeStart-radius),
                                                    end=(rangeEnd+radius))
        # Store the distance and element id of closest element that is not
        # overlapping the search range and is to the left of the search range
        less_than_closest_dist = radius
        less_than_closest_element = set()
        # Store the distance and element id of the closest element that is
        # not overlapping the search range and is to the right of the search range
        greater_than_closest_dist = radius
        greater_than_closest_element = set()
        # Store any overlapping elements
        overlapping = set()
        for theRange, element_id in iterator:
            # Skip if element_id not of the right type
            if self.graph.node[element_id]['type'] != element_type: continue
            # Close the Range if necessary
            if not all((theRange.isLowerBoundClosed(), theRange.isUpperBoundClosed())):
                newLowerBd = theRange.lowerEndpoint() if theRange.isLowerBoundClosed() else \
                  (theRange.lowerEndpoint()+1)
                newUpperBd = theRange.upperEndpoint() if theRange.isUpperBoundClosed() else \
                  (theRange.upperEndpoint()-1)
                theRange = Range.closed(newLowerBd, newUpperBd)
            if checkRange.isConnected(theRange):
                # Add element to overlapping set if overlapping
                overlapping.add(element_id)
            elif theRange.lowerEndpoint() < checkRange.lowerEndpoint():
                # If the element is to the left of the search range
                range_dist = checkRange.getDistanceFromRange(theRange)
                if range_dist < less_than_closest_dist:
                    # Add the element to the less than set after clearing current
                    # contents 
                    less_than_closest_element.clear()
                    less_than_closest_element.add(element_id)
                    less_than_closest_dist = range_dist
                elif range_dist == less_than_closest_dist:
                    # Add the element to the current set
                    less_than_closest_element.add(element_id)
            elif theRange.lowerEndpoint() > checkRange.upperEndpoint():
                # If the element is to the right of the search range
                if len(overlapping) > 0:
                    # Return the overlapping set if there are any overlapping elements
                    return overlapping
                range_dist = checkRange.getDistanceFromRange(theRange)
                if range_dist > less_than_closest_dist:
                    # Return the closest elements if the current range distance is
                    # greater than the left distances
                    if less_than_closest_dist < greater_than_closest_dist:
                        return less_than_closest_element
                    elif greater_than_closest_dist < less_than_closest_dist:
                        return greater_than_closest_element
                    else:
                        return less_than_closest_element.union(greater_than_closest_element)
                elif range_dist < greater_than_closest_dist:
                    # Add the element to the greater than set after clearing
                    # current contents
                    greater_than_closest_element.clear()
                    greater_than_closest_element.add(element_id)
                    greater_than_closest_dist = range_dist
                elif range_dist == greater_than_closest_dist:
                    # Add the element to the current set
                    greater_than_closest_element.add(element_id)
                elif range_dist > greater_than_closest_dist:
                    # Return the closest elements if the current range distance
                    # is greater than the minimum right distance
                    if less_than_closest_dist < greater_than_closest_dist:
                        return less_than_closest_element
                    elif greater_than_closest_dist < less_than_closest_dist:
                        return greater_than_closest_element
                    else:
                        return less_than_closest_element.union(greater_than_closest_element)
        # Return the closest if finished going through all elements
        if len(overlapping) > 0:
            return overlapping
        elif less_than_closest_dist < greater_than_closest_dist:
            return less_than_closest_element
        elif greater_than_closest_dist < less_than_closest_dist:
            return greater_than_closest_element
        elif less_than_closest_dist == greater_than_closest_dist:
            return less_than_closest_element.union(greater_than_closest_element)                    
    def get_element_children_ids(self, element_id):
        """ Gets the ids of the children of an element

        Parameters
        ----------
        element_id : str
            The element for which you want the children

        Raises
        ------
        KeyError
            If the element is not present

        Returns
        -------
        List of the children IDs of the element
        """
        return self.graph.successors(element_id)
    def get_element_parent_ids(self, element_id):
        """ Gets the ids of the parents of an element

        Parameters
        ----------
        element_id : str
            The element for which you want the parents

        Raises
        ------
        KeyError
            If the element is not present

        Returns
        -------
        List of the parent IDs of the element
        """
        return self.graph.predecessors(element_id)            
